// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.10
// 	protoc        v6.33.1
// source: pam.proto

package proto

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// state represents the stage object for the UI.
type Stage int32

const (
	// Stage_userSelection is to select a user.
	Stage_userSelection Stage = 0
	// Stage_brokerSelection is to select a broker.
	Stage_brokerSelection Stage = 1
	// Stage_authModeSelection is to select an authentication mode.
	Stage_authModeSelection Stage = 2
	// Stage_challenge lets the user enter a secret or wait for authentication from the broker.
	Stage_challenge Stage = 3
)

// Enum value maps for Stage.
var (
	Stage_name = map[int32]string{
		0: "userSelection",
		1: "brokerSelection",
		2: "authModeSelection",
		3: "challenge",
	}
	Stage_value = map[string]int32{
		"userSelection":     0,
		"brokerSelection":   1,
		"authModeSelection": 2,
		"challenge":         3,
	}
)

func (x Stage) Enum() *Stage {
	p := new(Stage)
	*p = x
	return p
}

func (x Stage) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Stage) Descriptor() protoreflect.EnumDescriptor {
	return file_pam_proto_enumTypes[0].Descriptor()
}

func (Stage) Type() protoreflect.EnumType {
	return &file_pam_proto_enumTypes[0]
}

func (x Stage) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Stage.Descriptor instead.
func (Stage) EnumDescriptor() ([]byte, []int) {
	return file_pam_proto_rawDescGZIP(), []int{0}
}

var File_pam_proto protoreflect.FileDescriptor

const file_pam_proto_rawDesc = "" +
	"\n" +
	"\tpam.proto\x12\x03pam*U\n" +
	"\x05Stage\x12\x11\n" +
	"\ruserSelection\x10\x00\x12\x13\n" +
	"\x0fbrokerSelection\x10\x01\x12\x15\n" +
	"\x11authModeSelection\x10\x02\x12\r\n" +
	"\tchallenge\x10\x03B/Z-github.com/canonical/authd/pam/internal/protob\x06proto3"

var (
	file_pam_proto_rawDescOnce sync.Once
	file_pam_proto_rawDescData []byte
)

func file_pam_proto_rawDescGZIP() []byte {
	file_pam_proto_rawDescOnce.Do(func() {
		file_pam_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_pam_proto_rawDesc), len(file_pam_proto_rawDesc)))
	})
	return file_pam_proto_rawDescData
}

var file_pam_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_pam_proto_goTypes = []any{
	(Stage)(0), // 0: pam.Stage
}
var file_pam_proto_depIdxs = []int32{
	0, // [0:0] is the sub-list for method output_type
	0, // [0:0] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_pam_proto_init() }
func file_pam_proto_init() {
	if File_pam_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_pam_proto_rawDesc), len(file_pam_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   0,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_pam_proto_goTypes,
		DependencyIndexes: file_pam_proto_depIdxs,
		EnumInfos:         file_pam_proto_enumTypes,
	}.Build()
	File_pam_proto = out.File
	file_pam_proto_goTypes = nil
	file_pam_proto_depIdxs = nil
}
